/*

Copyright (c) 2025 Д.С. Андреев

Этот файл — часть plik.

plik - свободная программа: вы можете перераспространять ее и/или
изменять ее на условиях Стандартной общественной лицензии GNU в том виде,
в каком она была опубликована Фондом свободного программного обеспечения;
либо версии 3 лицензии, либо (по вашему выбору) любой более поздней
версии.

plik распространяется в надежде, что она будет полезной,
но БЕЗО ВСЯКИХ ГАРАНТИЙ; даже без неявной гарантии ТОВАРНОГО ВИДА
или ПРИГОДНОСТИ ДЛЯ ОПРЕДЕЛЕННЫХ ЦЕЛЕЙ. Подробнее см. в Стандартной
общественной лицензии GNU.

Вы должны были получить копию Стандартной общественной лицензии GNU
вместе с этой программой. Если это не так, см. <http://www.gnu.org/licenses/>.


This file is part of plik.

plik is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

plik is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with plik. If not, see <http://www.gnu.org/licenses/>.

*/

/* функции работы с ОС и общие определения */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdarg.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <errno.h>

#define SIZE_OF(x) (sizeof(x)/sizeof(x[0]))

void console(const char * fmt,...) {
    time_t time_current = time(NULL);
    struct tm time_struct = * localtime(&time_current);

    char dt_buffer[256];
    strftime(dt_buffer,sizeof(dt_buffer),"%d.%m.%Y %H:%M:%S",&time_struct);

    va_list args;
    va_start(args,fmt);

    char text_buffer[2048];
    vsnprintf(text_buffer,sizeof(text_buffer),fmt,args);

    va_end(args);

    printf("[%s] %s\n",dt_buffer,text_buffer);
    }

void console_os_error(const char * message) {
    char * error_text = strerror(errno);
    console("%s: %s",message,error_text);
    }

/* autorelease pool для работы с памятью в C */

typedef struct {
    void * data_ptr;
    void (* release_func)(void *);
    } pool_element;

typedef struct {
    pool_element * element_ptr;
    int element_count;
    } pool_root;

pool_root * pool_init(pool_root * root) {
    root->element_ptr = NULL;
    root->element_count = 0;

    return root;
    }

void * pool_add(pool_root * root,void * data_ptr,void (* release_func)(void *)) {
    int count_old = root->element_count + 0;
    int count_new = root->element_count + 1;

    int size_old = count_old * sizeof(pool_element);
    int size_new = count_new * sizeof(pool_element);

    root->element_ptr = realloc(root->element_ptr,size_new);
    root->element_count = count_new;

    pool_element * new_element = root->element_ptr + count_old;
    new_element->data_ptr = data_ptr;
    new_element->release_func = release_func;

    return data_ptr;
    }

pool_root * pool_drain(pool_root * root) {
    for(int i = root->element_count - 1; i >= 0; i--) {
        pool_element * element = root->element_ptr + i;

        if(element->data_ptr == NULL) { continue; }
        if(element->release_func == NULL) { continue; }

        element->release_func(element->data_ptr);
        }

    free(root->element_ptr);
    pool_init(root);

    return root;
    }

/* бинарный кэш для хранения открытых файлов */

typedef struct {
    char * key_str;
    void * data_ptr;
    int data_l;
    } cache_element;

typedef struct {
    cache_element * element_ptr;
    int element_count;
    } cache_root;

cache_root * cache_init(cache_root * root) {
    root->element_ptr = NULL;
    root->element_count = 0;

    return root;
    }

bool cache_take(cache_root * root,const char * key_str,void * * data_ptr,int * data_l) {
    for(int i=0; i<root->element_count; i++) {
        cache_element * element = root->element_ptr + i;
        if(element->key_str == NULL) { continue; }
        if(strcmp(key_str,element->key_str) != 0) { continue; }

        free(element->key_str);
        element->key_str = NULL;

        * data_ptr = element->data_ptr;
        element->data_ptr = NULL;

        * data_l = element->data_l;
        element->data_l = 0;

        return true;
        }

    return false;
    }

cache_element * cache_get_unused_element(cache_root * root) {
    for(int i=0; i<root->element_count; i++) {
        cache_element * element = root->element_ptr + i;
        if(element->key_str != NULL) { continue; }

        return element;
        }

    int count_old = root->element_count + 0;
    int count_new = root->element_count + 1;

    int size_old = count_old * sizeof(cache_element);
    int size_new = count_new * sizeof(cache_element);

    root->element_ptr = realloc(root->element_ptr,size_new);
    root->element_count = count_new;

    cache_element * new_element = root->element_ptr + count_old;
    return new_element;
    }

cache_root * cache_put(cache_root * root,const char * key_str,void * data_ptr,int data_l) {
    cache_element * element = cache_get_unused_element(root);

    element->key_str = strdup(key_str);

    if(data_ptr != NULL && data_l > 0) {
        element->data_ptr = malloc(data_l);
        memcpy(element->data_ptr,data_ptr,data_l);

        element->data_l = data_l;
        }

    else {
        element->data_ptr = NULL;
        element->data_l = 0;
        }

    return root;
    }

cache_root * cache_drain(cache_root * root) {
    for(int i=0; i<root->element_count; i++) {
        cache_element * element = root->element_ptr + i;
        if(element->key_str == NULL) { continue; }

        free(element->key_str);
        free(element->data_ptr);
        }

    free(root->element_ptr);
    cache_init(root);

    return root;
    }

void cache_release(void * ptr) {
    cache_drain((cache_root *)ptr);
    }

#include <pthread.h>

pthread_mutex_t * mutex_init(pthread_mutex_t * mutex) {
    pthread_mutex_init(mutex,NULL);
    return mutex;
    }

void mutex_release(void * ptr) {
    pthread_mutex_destroy((pthread_mutex_t *)ptr);
    }

/* функции хэширования */

#include <openssl/evp.h>

bool calculate_md5(void * data,const int data_l,char * * md5,pool_root * pool) {
    EVP_MD_CTX * mdctx = EVP_MD_CTX_new();
    EVP_DigestInit_ex(mdctx,EVP_md5(),NULL);
    EVP_DigestUpdate(mdctx,data,data_l);

    unsigned int md5_digest_len = EVP_MD_size(EVP_md5());
    unsigned char * md5_digest = pool_add(pool,malloc(md5_digest_len),free);
    EVP_DigestFinal_ex(mdctx,md5_digest,&md5_digest_len);
    EVP_MD_CTX_free(mdctx);

    * md5 = pool_add(pool,malloc(md5_digest_len*2 + 1),free);

    for(int i=0; i<md5_digest_len; i++) {
        sprintf(&(* md5)[i*2],"%02x",(unsigned int)md5_digest[i]);
        }

    return true;
    }

/* функции компресии / декомпресии */

#include <zlib.h>

bool z_compress(const void * in_buffer,const int in_buffer_l,void * * out_buffer,int * out_buffer_l,pool_root * pool) {
    console("Сжатие блока длиной %d байт(а,ов)",in_buffer_l);

    z_stream strm;
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;

    if(deflateInit(&strm,Z_BEST_COMPRESSION) != Z_OK) {
        console("Ошибка инициализации сжатия: %s",strm.msg);
        return false;
        }

    * out_buffer_l = deflateBound(&strm,in_buffer_l);
    * out_buffer = pool_add(pool,malloc(* out_buffer_l),free);

    strm.avail_in = in_buffer_l;
    strm.next_in = (unsigned char *)in_buffer;
    strm.avail_out = * out_buffer_l;
    strm.next_out = * out_buffer;

    bool compress_done = true;

    if(deflate(&strm,Z_FINISH) != Z_STREAM_END) {
        console("Ошибка сжатия: %s",strm.msg);
        compress_done = false;
        }

    * out_buffer_l -= strm.avail_out;

    deflateEnd(&strm);
    return compress_done;
    }

bool z_decompress(void * in_buffer,const int in_buffer_l,void * * out_buffer,int * out_buffer_l,pool_root * pool) {
    console("Расжатие блока длиной %d байт(а,ов)",in_buffer_l);

    z_stream strm;
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;

    if(inflateInit(&strm) != Z_OK) {
        console("Ошибка инициализации расжатия: %s",strm.msg);
        return false;
        }

    void * temp_buffer = NULL;
    int temp_buffer_l = 0;

    strm.avail_in = in_buffer_l;
    strm.next_in = in_buffer;

    int inflate_done = Z_OK;

    do {
        if(strm.total_out >= temp_buffer_l) {
            temp_buffer_l += 1024;
            temp_buffer = realloc(temp_buffer,temp_buffer_l);
            }

        strm.avail_out = temp_buffer_l - strm.total_out;
        strm.next_out = temp_buffer + strm.total_out;

        inflate_done = inflate(&strm,Z_NO_FLUSH);
    } while(inflate_done == Z_OK);

    pool_add(pool,temp_buffer,free);

    * out_buffer_l = strm.total_out;
    * out_buffer = pool_add(pool,malloc(* out_buffer_l),free);

    memcpy(* out_buffer,temp_buffer,* out_buffer_l);

    if(inflate_done != Z_STREAM_END) {
        console("Ошибка расжатия: %s",strm.msg);
        }

    inflateEnd(&strm);
    return ( inflate_done == Z_STREAM_END ? true : false );
    }

/* функции работы с БД */

#include <sqlite3.h>

sqlite3 * g_db = NULL;

bool db_raw(const char * raw_request,int * changes_count) {
    bool exec_ok = true;
    char * err_msg = NULL;

    if(sqlite3_exec(g_db,raw_request,NULL,NULL,&err_msg) != SQLITE_OK) {
        console("Ошибка запроса '%s': %s",raw_request,err_msg);
        exec_ok = false;
        }

    else { if(changes_count != NULL) {
        * changes_count = sqlite3_changes(g_db);
        } }

    if(err_msg != NULL) { sqlite3_free(err_msg); }
    return exec_ok;
    }

void db_disconnect(void * ptr) {
    sqlite3 * db = (sqlite3 *)ptr;

    if(sqlite3_close(db) == SQLITE_OK) { return; }
    else { console("Ошибка закрытия: %s",sqlite3_errmsg(db)); }
    }

bool db_connect(const char * db_name,pool_root * pool) {
    if(sqlite3_open_v2(db_name,&g_db,SQLITE_OPEN_READWRITE|SQLITE_OPEN_CREATE|SQLITE_OPEN_FULLMUTEX,NULL) != SQLITE_OK) {
        console("Ошибка открытия '%s': %s",db_name,sqlite3_errmsg(g_db));
        return false;
        }

    pool_add(pool,g_db,db_disconnect);

    const char * requests[] = {
        "PRAGMA journal_mode = OFF",
        "PRAGMA synchronous = OFF",
        "PRAGMA cache_size = -65536",
        "PRAGMA locking_mode = EXCLUSIVE",
        "PRAGMA temp_store = MEMORY",

        "CREATE TABLE IF NOT EXISTS t_entry ("
        "  entry_id INTEGER PRIMARY KEY AUTOINCREMENT,"
        "  entry_parent_id INTEGER NOT NULL DEFAULT 0,"
        "  entry_name TEXT NOT NULL,"
        "  entry_folder_file INTEGER(1) NOT NULL DEFAULT 0,"
        "  entry_block_id TEXT,"
        "  entry_ctime BIGINT NOT NULL DEFAULT 0,"
        "  entry_mtime BIGINT NOT NULL DEFAULT 0"
        "  )",

        "CREATE INDEX IF NOT EXISTS i_entry_parent_id ON t_entry(entry_parent_id)",
        "CREATE INDEX IF NOT EXISTS i_entry_block_id ON t_entry(entry_block_id)",
        "CREATE INDEX IF NOT EXISTS i_entry_parent_id_entry_name ON t_entry(entry_parent_id,entry_name)",

        "CREATE TRIGGER IF NOT EXISTS r_entry_insert AFTER INSERT ON t_entry"
        "  BEGIN"
        "    UPDATE t_entry SET entry_ctime=strftime('%s','now') WHERE rowid=new.rowid;"
        "  END",

        "CREATE TRIGGER IF NOT EXISTS r_entry_update AFTER UPDATE ON t_entry"
        "  BEGIN"
        "    UPDATE t_entry SET entry_mtime=strftime('%s','now') WHERE rowid=NEW.rowid;"
        "  END",

        "ANALYZE t_entry",

        "CREATE TABLE IF NOT EXISTS t_block ("
        "  block_id TEXT PRIMARY KEY UNIQUE,"
        "  block_content BLOB NOT NULL,"
        "  block_size BIGINT NOT NULL"
        "  )",

        "ANALYZE t_block",

        "CREATE VIEW IF NOT EXISTS v_entry AS"
        "  SELECT *"
        "  ,COALESCE(t_block.block_size,0) AS entry_block_size"
        "  ,IIF(entry_parent_id=0,1,("
        "    SELECT COUNT(entry_id)"
        "    FROM t_entry t_parent"
        "    WHERE t_parent.entry_id=t_entry.entry_parent_id"
        "  ) ) AS entry_parent_count"
        "  FROM t_entry"
        "  LEFT JOIN t_block ON t_entry.entry_block_id=t_block.block_id",

        "CREATE VIEW IF NOT EXISTS v_block AS"
        "  SELECT *"
        "  ,("
        "    SELECT COUNT(entry_id)"
        "    FROM t_entry"
        "    WHERE t_entry.entry_block_id=t_block.block_id"
        "  ) AS block_entry_count"
        "  ,IIF("
        "    block_size > 0,"
        "    CAST(LENGTH(block_content) AS REAL)/block_size,"
        "    NULL"
        "  ) AS block_compress_ratio"
        "  FROM t_block",
        };

    for(int i=0; i<SIZE_OF(requests); i++) {
        if(db_raw(requests[i],NULL) == false) { return false; }
        }

    return true;
    }

bool db_cleanup(void) {
    const char * requests[] = {
        "DELETE FROM t_entry "
        "WHERE entry_id IN ("
        "  SELECT entry_id"
        "  FROM v_entry"
        "  WHERE entry_parent_count=0"
        "  )",

        "DELETE FROM t_block "
        "WHERE block_id IN ("
        "  SELECT block_id"
        "  FROM v_block"
        "  WHERE block_entry_count=0"
        "  )",
        };

    int changed_count = 0;

    for(int i=0; i<SIZE_OF(requests); i++) {
        do { if(db_raw(requests[i],&changed_count) == false) { return false; } } while(changed_count);
        }

    return true;
    }

void m_finalize_stmt(void * ptr) {
    sqlite3_stmt * stmt = (sqlite3_stmt *)ptr;
    sqlite3_finalize(stmt);
    }

/* функции для работы с элементами (файлами/каталогами) внутри БД */

bool m_get_entry_by_parent_name(const char * token,int * entry_id,pool_root * pool) {
    console("Поиск элемента '%s' с родителем %d",token,* entry_id);
    bool entry_found = false;

    const char * select_request =
        "SELECT entry_id "
        "FROM t_entry "
        "WHERE entry_parent_id=:entry_parent_id "
        "AND entry_name=:entry_name";

    sqlite3_stmt * select_query = NULL;
    int select_prepared = sqlite3_prepare_v2(g_db,select_request,-1,&select_query,NULL);
    pool_add(pool,select_query,m_finalize_stmt);

    if(select_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",select_request,sqlite3_errmsg(g_db));
        return entry_found;
        }

    sqlite3_bind_int(select_query,sqlite3_bind_parameter_index(select_query,":entry_parent_id"),* entry_id);
    sqlite3_bind_text(select_query,sqlite3_bind_parameter_index(select_query,":entry_name"),token,-1,SQLITE_STATIC);

    while(sqlite3_step(select_query) == SQLITE_ROW) {
        * entry_id = sqlite3_column_int(select_query,0);
        entry_found = true;
        }

    return entry_found;
    }

bool m_get_entry_by_path(const char * path_orig,int * entry_id,pool_root * pool) {
    console("Поиск пути '%s' с родителем %d",path_orig,* entry_id);
    char * path = pool_add(pool,strdup(path_orig),free);

    bool entry_found = true;

    char * save_ptr = NULL;
    char * token = strtok_r(path,"/",&save_ptr);

    while(token != NULL) {
        if(m_get_entry_by_parent_name(token,entry_id,pool) == false ) { return false; }
        token = strtok_r(NULL,"/",&save_ptr);
        }

    return true;
    }

bool m_make_entry_parent(const char * token,int * entry_id,const bool folder_file,pool_root * pool) {
    console("Создание элемента %s '%s' с родителем %d",(folder_file ? "каталог" : "файл"),token,* entry_id);

    const char * insert_request =
        "INSERT INTO t_entry(entry_parent_id,entry_name,entry_folder_file) "
        "VALUES(:entry_parent_id,:entry_name,:entry_folder_file) "
        "RETURNING entry_id";

    sqlite3_stmt * insert_query = NULL;
    int insert_prepared = sqlite3_prepare_v2(g_db,insert_request,-1,&insert_query,NULL);
    pool_add(pool,insert_query,m_finalize_stmt);

    if(insert_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",insert_request,sqlite3_errmsg(g_db));
        return false;
        }

    sqlite3_bind_int(insert_query,sqlite3_bind_parameter_index(insert_query,":entry_parent_id"),* entry_id);
    sqlite3_bind_text(insert_query,sqlite3_bind_parameter_index(insert_query,":entry_name"),token,-1,SQLITE_STATIC);
    sqlite3_bind_int(insert_query,sqlite3_bind_parameter_index(insert_query,":entry_folder_file"),(folder_file ? 1 : 0));

    bool entry_created = false;

    while(sqlite3_step(insert_query) == SQLITE_ROW) {
        * entry_id = sqlite3_column_int(insert_query,0);
        entry_created = true;
        }

    if(entry_created == false) { console("Ошибка создания элемента"); }

    return entry_created;
    }

bool m_make_path(const char * path_orig,const bool folder_file,pool_root * pool) {
    console("Создание пути '%s' как %s",path_orig,( folder_file == true ? "каталог" : "файл"));
    char * path = pool_add(pool,strdup(path_orig),free);

    int entry_id = 0;

    char * save_ptr = NULL;
    char * token = strtok_r(path,"/",&save_ptr);

    while(token != NULL) {
        if(m_get_entry_by_parent_name(token,&entry_id,pool) == false) {
            return m_make_entry_parent(token,&entry_id,folder_file,pool);
            }

        token = strtok_r(NULL,"/",&save_ptr);
        }

    return true;
    }

bool m_remove_entry(const int entry_id,pool_root * pool) {
    console("Удаление элемента %d",entry_id);

    const char * delete_request =
        "DELETE FROM t_entry "
        "WHERE entry_id=:entry_id";

    sqlite3_stmt * delete_query = NULL;
    int delete_prepared = sqlite3_prepare_v2(g_db,delete_request,-1,&delete_query,NULL);
    pool_add(pool,delete_query,m_finalize_stmt);

    if(delete_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",delete_request,sqlite3_errmsg(g_db));
        return false;
        }

    sqlite3_bind_int(delete_query,sqlite3_bind_parameter_index(delete_query,":entry_id"),entry_id);

    if(sqlite3_step(delete_query) != SQLITE_DONE) {
        console("Ошибка выполнения запроса '%s': %s",delete_request,sqlite3_errmsg(g_db));
        return false;
        }

    return true;
    }

bool m_remove_path(const char * path_orig,pool_root * pool) {
    console("Удаление пути '%s'",path_orig);
    char * path = pool_add(pool,strdup(path_orig),free);

    int entry_id = 0;

    char * save_ptr = NULL;
    char * token = strtok_r(path,"/",&save_ptr);

    while(token != NULL) {
        if(m_get_entry_by_parent_name(token,&entry_id,pool) == false) { return false; }
        token = strtok_r(NULL,"/",&save_ptr);
        }

    if(m_remove_entry(entry_id,pool) == false) { return false; }

    return true;
    }

bool m_set_entry_block(const int entry_id,const char * hash,pool_root * pool) {
    console("Обновление элемента %d блоком '%s'",entry_id,hash);

    const char * update_request =
        "UPDATE t_entry "
        "SET entry_block_id=:entry_block_id "
        "WHERE entry_id=:entry_id";

    sqlite3_stmt * update_query = NULL;
    int update_prepared = sqlite3_prepare_v2(g_db,update_request,-1,&update_query,NULL);
    pool_add(pool,update_query,m_finalize_stmt);

    if(update_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",update_request,sqlite3_errmsg(g_db));
        return false;
        }

    sqlite3_bind_text(update_query,sqlite3_bind_parameter_index(update_query,":entry_block_id"),hash,-1,SQLITE_STATIC);
    sqlite3_bind_int(update_query,sqlite3_bind_parameter_index(update_query,":entry_id"),entry_id);

    if(sqlite3_step(update_query) != SQLITE_DONE) {
        console("Ошибка выполнения запроса '%s': %s",update_request,sqlite3_errmsg(g_db));
        return false;
        }

    return true;
    }

bool m_set_entry_parent_name(const int entry_id,const int parent_id, const char * name,pool_root * pool) {
    console("Обновление элемента %d родителем %d названием '%s'",entry_id,parent_id,name);

    const char * update_request =
        "UPDATE t_entry "
        "SET entry_parent_id=:entry_parent_id,entry_name=:entry_name "
        "WHERE entry_id=:entry_id";

    sqlite3_stmt * update_query = NULL;
    int update_prepared = sqlite3_prepare_v2(g_db,update_request,-1,&update_query,NULL);
    pool_add(pool,update_query,m_finalize_stmt);

    if(update_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",update_request,sqlite3_errmsg(g_db));
        return false;
        }

    sqlite3_bind_int(update_query,sqlite3_bind_parameter_index(update_query,":entry_parent_id"),parent_id);
    sqlite3_bind_text(update_query,sqlite3_bind_parameter_index(update_query,":entry_name"),name,-1,SQLITE_STATIC);
    sqlite3_bind_int(update_query,sqlite3_bind_parameter_index(update_query,":entry_id"),entry_id);

    if(sqlite3_step(update_query) != SQLITE_DONE) {
        console("Ошибка выполнения запроса '%s': %s",update_request,sqlite3_errmsg(g_db));
        return false;
        }

    return true;
    }

bool m_rename_path(const char * from_path,const char * to_path_orig,pool_root * pool) {
    console("Переименование '%s' в '%s'",from_path,to_path_orig);
    char * to_path = pool_add(pool,strdup(to_path_orig),free);

    int from_entry_id = 0;
    if(m_get_entry_by_path(from_path,&from_entry_id,pool) == false) { return false; }

    int to_entry_id = 0;
    int to_parent_id = 0;
    char * to_token = NULL;

    char * save_ptr = NULL;
    char * token = strtok_r(to_path,"/",&save_ptr);

    if(token == NULL) {
        console("Пустой целевой путь (не найдено ни одного элемента)");
        return false;
        }

    while(token != NULL) {
        to_parent_id = to_entry_id;
        to_token = pool_add(pool,strdup(token),free);

        if(m_get_entry_by_parent_name(token,&to_entry_id,pool) == false) { break; }

        token = strtok_r(NULL,"/",&save_ptr);
        }

    if(m_set_entry_parent_name(from_entry_id,to_parent_id,to_token,pool) == false) { return false; }

    return true;
    }

/* функции для работы с блоками внутри БД */

bool m_get_entry_block(const int entry_id,void * * buffer,int * buffer_l,pool_root * pool) {
    console("Получение блока для элемента %d",entry_id);

    const char * select_request =
        "SELECT block_content "
        "FROM v_entry "
        "WHERE entry_id=:entry_id";

    sqlite3_stmt * select_query = NULL;
    int select_prepared = sqlite3_prepare_v2(g_db,select_request,-1,&select_query,NULL);
    pool_add(pool,select_query,m_finalize_stmt);

    if(select_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",select_request,sqlite3_errmsg(g_db));
        return false;
        }

    sqlite3_bind_int(select_query,sqlite3_bind_parameter_index(select_query,":entry_id"),entry_id);

    void * compressed = NULL;
    int compressed_l = 0;

    while(sqlite3_step(select_query) == SQLITE_ROW) {
        if(compressed != NULL) { continue; }

        compressed_l = sqlite3_column_bytes(select_query,0);
        compressed = pool_add(pool,malloc(compressed_l),free);

        memcpy(compressed,sqlite3_column_blob(select_query,0),compressed_l);
        }

    if(compressed != NULL && compressed_l > 0) {
        if(z_decompress(compressed,compressed_l,buffer,buffer_l,pool) == false) { return false; }
        return true;
        }

    * buffer = NULL;
    * buffer_l = 0;

    return true;
    }

bool m_count_blocks(const char * hash,int * existing_count,pool_root * pool) {
    console("Подсчёт блоков '%s'",hash);
    bool entry_found = false;

    const char * select_request =
        "SELECT COUNT(block_id) AS result "
        "FROM t_block "
        "WHERE block_id=:block_id";

    sqlite3_stmt * select_query = NULL;
    int select_prepared = sqlite3_prepare_v2(g_db,select_request,-1,&select_query,NULL);
    pool_add(pool,select_query,m_finalize_stmt);

    if(select_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",select_request,sqlite3_errmsg(g_db));
        return entry_found;
        }

    sqlite3_bind_text(select_query,sqlite3_bind_parameter_index(select_query,":block_id"),hash,-1,SQLITE_STATIC);

    while(sqlite3_step(select_query) == SQLITE_ROW) {
        * existing_count = sqlite3_column_int(select_query,0);
        entry_found = true;
        }

    return entry_found;
    }

bool m_put_block(const char * hash,const void * buffer,const int buffer_l,pool_root * pool) {
    console("Сохранение блока '%s' длины %d",hash,buffer_l);

    void * compressed = NULL; int compressed_l = 0;
    if(z_compress(buffer,buffer_l,&compressed,&compressed_l,pool) == false) { return false; }

    const char * insert_request =
        "INSERT INTO t_block(block_id,block_size,block_content) "
        "VALUES(:block_id,:block_size,:block_content)";

    sqlite3_stmt * insert_query = NULL;
    int insert_prepared = sqlite3_prepare_v2(g_db,insert_request,-1,&insert_query,NULL);
    pool_add(pool,insert_query,m_finalize_stmt);

    if(insert_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",insert_request,sqlite3_errmsg(g_db));
        return false;
        }

    sqlite3_bind_text(insert_query,sqlite3_bind_parameter_index(insert_query,":block_id"),hash,-1,SQLITE_STATIC);
    sqlite3_bind_int(insert_query,sqlite3_bind_parameter_index(insert_query,":block_size"),buffer_l);
    sqlite3_bind_blob(insert_query,sqlite3_bind_parameter_index(insert_query,":block_content"),compressed,compressed_l,SQLITE_STATIC);

    if(sqlite3_step(insert_query) != SQLITE_DONE) {
        console("Ошибка выполнения запроса '%s': %s",insert_request,sqlite3_errmsg(g_db));
        return false;
        }

    return true;
    }

/* объявления кэша и мутекса к нему */

cache_root g_cache;
pthread_mutex_t g_cache_mutex;

/* функции интерфейса к fuse */

#define FUSE_USE_VERSION 26
#include <fuse.h>

bool p_getattr(const char * path, struct stat * stbuf,pool_root * pool) {
    console("Получение атрибутов '%s'",path);

    if(strcmp(path,"/") == 0) {
        stbuf->st_mode = (S_IFDIR|0755);
        stbuf->st_size = 0;
        stbuf->st_ctim.tv_sec = time(NULL);
        stbuf->st_mtim.tv_sec = time(NULL);
        stbuf->st_uid = getuid();
        stbuf->st_gid = getgid();

        return true;
        }

    int entry_id = 0;
    if(m_get_entry_by_path(path,&entry_id,pool) == false) { return false; }

    bool entry_found = false;

    const char * select_request =
        "SELECT entry_folder_file "
        ",entry_block_size "
        ",entry_ctime "
        ",entry_mtime "
        "FROM v_entry "
        "WHERE entry_id=:entry_id";

    sqlite3_stmt * select_query = NULL;
    int select_prepared = sqlite3_prepare_v2(g_db,select_request,-1,&select_query,NULL);
    pool_add(pool,select_query,m_finalize_stmt);

    if(select_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",select_request,sqlite3_errmsg(g_db));
        return false;
        }

    sqlite3_bind_int(select_query,sqlite3_bind_parameter_index(select_query,":entry_id"),entry_id);

    while(sqlite3_step(select_query) == SQLITE_ROW) {
        stbuf->st_mode = ( sqlite3_column_int(select_query,0) ? (S_IFDIR|0755) : (S_IFREG|0644) );
        stbuf->st_size = sqlite3_column_int(select_query,1);
        stbuf->st_ctim.tv_sec = sqlite3_column_int(select_query,2);
        stbuf->st_mtim.tv_sec = sqlite3_column_int(select_query,3);
        stbuf->st_uid = getuid();
        stbuf->st_gid = getgid();

        entry_found = true;
        }

    return entry_found;
    }

static int g_getattr(const char * path, struct stat * stbuf) {
    pool_root pool;
    pool_init(&pool);

    bool getattr_rc = p_getattr(path,stbuf,&pool);

    pool_drain(&pool);
    return ( getattr_rc == true ? 0 : -ENOENT );
    }

bool p_readdir(const char * path, void * buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info * fi,pool_root * pool) {
    console("Чтение каталога '%s' смещение %d",path,offset);

    int entry_id = 0;
    if(m_get_entry_by_path(path,&entry_id,pool) == false) { return false; }

    filler(buf,".",NULL,0);
    filler(buf,"..",NULL,0);

    const char * select_request =
        "SELECT entry_name "
        "FROM t_entry "
        "WHERE entry_parent_id=:entry_parent_id";

    sqlite3_stmt * select_query = NULL;
    int select_prepared = sqlite3_prepare_v2(g_db,select_request,-1,&select_query,NULL);
    pool_add(pool,select_query,m_finalize_stmt);

    if(select_prepared != SQLITE_OK) {
        console("Ошибка подготовки запроса '%s': %s",select_request,sqlite3_errmsg(g_db));
        return false;
        }

    sqlite3_bind_int(select_query,sqlite3_bind_parameter_index(select_query,":entry_parent_id"),entry_id);

    while(sqlite3_step(select_query) == SQLITE_ROW) {
        filler(buf,sqlite3_column_text(select_query,0),NULL,0);
        }

    return true;
    }

static int g_readdir(const char * path, void * buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info * fi) {
    pool_root pool;
    pool_init(&pool);

    bool readdir_rc = p_readdir(path,buf,filler,offset,fi,&pool);

    pool_drain(&pool);
    return ( readdir_rc == true ? 0 : -ENOENT );
    }

bool m_open(const char * path, struct fuse_file_info * fi,pool_root * pool) {
    console("Открытие '%s'",path);

    int entry_id = 0;
    if(m_get_entry_by_path(path,&entry_id,pool) == false) { return false; }

    void * buffer; int buffer_l;
    if(m_get_entry_block(entry_id,&buffer,&buffer_l,pool) == false) { return false; }

    pthread_mutex_lock(&g_cache_mutex);
    cache_put(&g_cache,path,buffer,buffer_l);
    pthread_mutex_unlock(&g_cache_mutex);

    return true;
    }

static int g_open(const char * path, struct fuse_file_info * fi) {
    pool_root pool;
    pool_init(&pool);

    bool open_rc = m_open(path,fi,&pool);

    pool_drain(&pool);
    return ( open_rc == true ? 0 : -ENOENT );
    }

bool m_release(const char * path, struct fuse_file_info * fi,pool_root * pool) {
    console("Закрытие '%s'",path);

    int entry_id = 0;
    if(m_get_entry_by_path(path,&entry_id,pool) == false) { return false; }

    pthread_mutex_lock(&g_cache_mutex);

    void * buffer; int buffer_l;
    bool get_ok = cache_take(&g_cache,path,&buffer,&buffer_l);

    pthread_mutex_unlock(&g_cache_mutex);

    if(get_ok == false) {
        console("Закрытие не открытого файла '%s'",path);
        return -ENOENT;
        }

    pool_add(pool,buffer,free);

    char * md5 = NULL;
    if(calculate_md5(buffer,buffer_l,&md5,pool) == false) { return false; }

    int existing_count = 0;
    if(m_count_blocks(md5,&existing_count,pool) == false) { return false; }

    if(existing_count < 1) {
        if(m_put_block(md5,buffer,buffer_l,pool) == false) { return false; }
        }

    if(m_set_entry_block(entry_id,md5,pool) == false) { return false; }

    return true;
    }

static int g_release(const char * path, struct fuse_file_info * fi) {
    pool_root pool;
    pool_init(&pool);

    bool release_rc = m_release(path,fi,&pool);

    pool_drain(&pool);
    return ( release_rc == true ? 0 : -ENOENT );
    }

int p_read(const char * path, char * buf, size_t size, off_t offset, struct fuse_file_info * fi,pool_root * pool) {
    console("Чтение '%s' размер %d смещение %d",path,size,offset);

    pthread_mutex_lock(&g_cache_mutex);

    void * buffer; int buffer_l;
    bool get_ok = cache_take(&g_cache,path,&buffer,&buffer_l);

    pthread_mutex_unlock(&g_cache_mutex);

    if(get_ok == false) {
        console("Чтение из неоткрытого файла '%s'",path);
        return -ENOENT;
        }

    pool_add(pool,buffer,free);

    int read_l = ( (offset+size) > buffer_l ? (buffer_l-offset) : size );
    if(read_l < 0) { read_l = 0; }

    memcpy(buf,buffer+offset,read_l);

    pthread_mutex_lock(&g_cache_mutex);
    cache_put(&g_cache,path,buffer,buffer_l);
    pthread_mutex_unlock(&g_cache_mutex);

    return read_l;
    }

static int g_read(const char * path, char * buf, size_t size, off_t offset, struct fuse_file_info * fi) {
    pool_root pool;
    pool_init(&pool);

    int read_rc = p_read(path,buf,size,offset,fi,&pool);

    pool_drain(&pool);
    return read_rc;
    }

static int g_mknod(const char * path, mode_t mode, dev_t dev) {
    pool_root pool;
    pool_init(&pool);

    bool make_rc = m_make_path(path,false,&pool);

    pool_drain(&pool);
    return ( make_rc == true ? 0 : -EACCES );
    }

static int g_mkdir(const char * path, mode_t mode) {
    pool_root pool;
    pool_init(&pool);

    bool make_rc = m_make_path(path,true,&pool);

    pool_drain(&pool);
    return ( make_rc == true ? 0 : -EACCES );
    }

bool p_truncate(const char * path, off_t off, pool_root * pool) {
    console("Обрезание '%s' смещение %d",path,off);

    pthread_mutex_lock(&g_cache_mutex);

    void * buffer; int buffer_l;
    bool get_ok = cache_take(&g_cache,path,&buffer,&buffer_l);

    pthread_mutex_unlock(&g_cache_mutex);

    if(get_ok == false) {
        console("Обрезание не открытого файла '%s'",path);
        return -ENOENT;
        }

    if(off != buffer_l) {
        buffer = realloc(buffer,off);
        buffer_l = off;
        }

    pool_add(pool,buffer,free);

    pthread_mutex_lock(&g_cache_mutex);
    cache_put(&g_cache,path,buffer,buffer_l);
    pthread_mutex_unlock(&g_cache_mutex);

    return true;
    }

static int g_truncate(const char * path, off_t off) {
    pool_root pool;
    pool_init(&pool);

    bool truncate_rc = p_truncate(path,true,&pool);

    pool_drain(&pool);
    return ( truncate_rc == true ? 0 : -EACCES );
    }

int p_write(const char * path, const char * buf, size_t size, off_t offset, struct fuse_file_info * fi,pool_root * pool) {
    console("Запись '%s' размер %d смещение %d",path,size,offset);

    pthread_mutex_lock(&g_cache_mutex);

    void * buffer; int buffer_l;
    bool get_ok = cache_take(&g_cache,path,&buffer,&buffer_l);

    pthread_mutex_unlock(&g_cache_mutex);

    if(get_ok == false) {
        console("Запись в неоткрытый файл '%s'",path);
        return -ENOENT;
        }

    int new_size = offset + size;

    if(new_size > buffer_l) {
        buffer = realloc(buffer,new_size);
        buffer_l = new_size;
        }

    pool_add(pool,buffer,free);

    memcpy(buffer + offset,buf,size);

    pthread_mutex_lock(&g_cache_mutex);
    cache_put(&g_cache,path,buffer,buffer_l);
    pthread_mutex_unlock(&g_cache_mutex);

    return size;
    }

static int g_write(const char * path, const char * buf, size_t size, off_t offset, struct fuse_file_info * fi) {
    pool_root pool;
    pool_init(&pool);

    int write_rc = p_write(path,buf,size,offset,fi,&pool);

    pool_drain(&pool);
    return write_rc;
    }

static int g_unlink(const char * path) {
    pool_root pool;
    pool_init(&pool);

    bool remove_rc = m_remove_path(path,&pool);

    pool_drain(&pool);
    return ( remove_rc == true ? 0 : -EACCES );
    }

static int g_rmdir(const char * path) {
    pool_root pool;
    pool_init(&pool);

    bool remove_rc = m_remove_path(path,&pool);

    pool_drain(&pool);
    return ( remove_rc == true ? 0 : -EACCES );
    }

static int g_rename(const char * from_path, const char * to_path) {
    pool_root pool;
    pool_init(&pool);

    bool rename_rc = m_rename_path(from_path,to_path,&pool);

    pool_drain(&pool);
    return ( rename_rc == true ? 0 : -EACCES );
    }

/* кровавая текущая муть */

struct plik_config {
    char * db_name;
    };

enum {
    KEY_HELP,
    };

#define PLIK_OPT(t,p,v) { t,offsetof(struct plik_config,p),v }

static struct fuse_opt plik_opts[] = {
    PLIK_OPT("db_name=%s",db_name,0),

    FUSE_OPT_KEY("-h",KEY_HELP),
    FUSE_OPT_KEY("--help",KEY_HELP),

    FUSE_OPT_END
    };

static int plik_opt_proc(void * data,const char * arg,int key,struct fuse_args * outargs) {
    switch (key) {
        case KEY_HELP:
            fprintf(stderr,"plik options:\n    -o db_name=STRING      database name to use (:memory: for ramdisk)\n\n");

            static struct fuse_operations operations;
            memset(&operations,0,sizeof(operations));

            fuse_opt_add_arg(outargs,"-h");
            fuse_main(outargs->argc,outargs->argv,&operations,NULL);

            exit(1);
        }

    return 1;
    }

/* точка входа */

int p_main(int argc,char * argv[],pool_root * pool) {
    struct fuse_args args = FUSE_ARGS_INIT(argc,argv);

    struct plik_config config;
    memset(&config,0,sizeof(config));

    fuse_opt_parse(&args,&config,plik_opts,plik_opt_proc);

    if(config.db_name == NULL) {
        console("Database not specified (append -o db_name=<db>)");
        return EXIT_FAILURE;
        }

    if(db_connect(config.db_name,pool) == false) { return false; }
    if(db_cleanup() == false) { return false; }

    pool_add(pool,cache_init(&g_cache),cache_release);
    pool_add(pool,mutex_init(&g_cache_mutex),mutex_release);

    static struct fuse_operations operations;
    memset(&operations,0,sizeof(operations));

    operations.getattr = g_getattr;
    operations.readdir = g_readdir;
    operations.open = g_open;
    operations.release = g_release;
    operations.read = g_read;
    operations.mknod = g_mknod;
    operations.mkdir = g_mkdir;
    operations.truncate = g_truncate;
    operations.write = g_write;
    operations.unlink = g_unlink;
    operations.rmdir = g_rmdir;
    operations.rename = g_rename;

    return fuse_main(args.argc,args.argv,&operations,NULL);
    }

int main(int argc,char * argv[]) {
    pool_root pool;
    pool_init(&pool);

    int main_rc = p_main(argc,argv,&pool);

    pool_drain(&pool);
    return main_rc;
    }
